#include <stdio.h>
#include <math.h>
#include <sys/time.h>

/*
#define TESTING
*/

#ifdef TESTING
static int numoftrials = 1; /*300*/;
#else
static int numoftrials = 100; /*300*/;
#endif

#ifdef TESTING
#define PRECISION 4
#else
#define PRECISION 256
#endif
#define COARSENESS 0x8000
#define COARSENESS_SHIFTED 0x80000000
typedef struct {
  int lower;
  int upper;
} tempevent_type;

typedef struct {
  int left[PRECISION+2];
  int right[PRECISION+2];
} tempdist_type;

typedef enum {FALSE, TRUE} bool;

tempdist_type uniform;

int maxmaxqueue = 0;
int maxqueue = 0;

int pregammas[101];

typedef struct {
  unsigned int leftrods;
  unsigned int rightrods;
  void *left;
  void *right;
} tree_node_type;

tree_node_type spare_nodes[1000];
int allocated = 0;
tree_node_type *uniform_tree;

typedef struct {
  unsigned int agranuleslower;
  unsigned int agranulesupper;
  tree_node_type *arodstree;
  unsigned int bgranuleslower;
  unsigned int bgranulesupper;
  tree_node_type *brodstree;
  /*
  bool         agranuleslowerboundary;
  bool         agranulesupperboundary;
  bool         bgranuleslowerboundary;
  bool         bgranulesupperboundary;
  */
} queue_node_type;

queue_node_type queue[8000];

tree_node_type *build_uniform_tree(int current) {
  tree_node_type *node;

  node = &spare_nodes[allocated++];
  if (current == 1) {
    node->left = NULL;
    node->right = NULL;
    node->leftrods = 0;
    node->rightrods = 0;
    }
  else {
    node->left = (void *)build_uniform_tree(current >> 1);
    node->right = (void *)build_uniform_tree(current >> 1);
    node->leftrods = current >> 1;
    node->rightrods = current >> 1;
    }
    return node;
 }

void main() {
  tree_node_type *build_uniform_tree();
  int i, j, k;
  int times[101];
  int temps[101];
  void printf(), gettimeofday();
  bool before();
  struct timeval tp1, tp2;
  int temp1, temp2;
  int gamma, temp;
  tempevent_type alpha, beta;
  double f;

  temp = 0;
  alpha.lower = 1;
  alpha.upper = 1023;
  beta.lower = 1;
  beta.upper = 1023; 
  for (i=1;i<PRECISION+1;i++) {
    uniform.left[i] = ((COARSENESS/PRECISION) * i);
    /*
    uniform.right[i-1] = ((COARSENESS/PRECISION) * i);
    */
    uniform.right[i] = ((COARSENESS/PRECISION) * i);
    }
  uniform.left[0] = 0;
  uniform.right[0] = 0;
  uniform.left[PRECISION+1] = 2*COARSENESS;
  uniform.right[PRECISION+1] = 2*COARSENESS;
  /*
  uniform.right[PRECISION] = PRECISION;
  */

#define GAMMA_MULT PRECISION * PRECISION / 100.0
  for (i=0;i<101;i++) {
    f = i;
    pregammas[i] = (int) ceil(f * GAMMA_MULT);
    /* printf("%d %d %f \n", i, pregammas[i], (i * (GAMMA_MULT)));*/
    }

  uniform_tree = build_uniform_tree(PRECISION);

  for (i=1+numoftrials;i<=100+numoftrials;i++) {

    gettimeofday(&tp1,NULL);

    for (j=0; j<numoftrials; j++) temp += before(i - numoftrials,&alpha,&beta);

    gettimeofday(&tp2,NULL);

    temp1 = tp1.tv_sec * 1000000 + tp1.tv_usec;
    temp2 = tp2.tv_sec * 1000000 + tp2.tv_usec;
    temp1 = temp2 - temp1;

    /*
    if (k) printf("TRUE %d %d ", k, i);
    else printf("False %d %d ", k, i);
    printf("%d\n", (temp1/numoftrials));
    */
    times[i-numoftrials] = temp1;
    temps[i-numoftrials] = temp;
    }
    for (i = 1; i <= 100; i++) printf("%d %d %d\n", temps[i], i, times[i]/numoftrials);
   printf("m %d\n", maxmaxqueue);
}

#define PUSH(agl,agu,ar,bgl,bgu,br) \
  if (ar && br) {\
    queue_ptr = &queue[maxqueue++]; \
    queue_ptr->agranuleslower = agl; \
    queue_ptr->agranulesupper = agu; \
    queue_ptr->arodstree = ar; \
    queue_ptr->bgranuleslower = bgl; \
    queue_ptr->bgranulesupper = bgu; \
    queue_ptr->brodstree = br; \
    }

#define POP(agl,agu,ar,bgl,bgu,br) \
 if (currentqueue == maxqueue) return FALSE; \
 else { \
  queue_ptr = &queue[currentqueue++]; \
  agl = queue_ptr->agranuleslower; \
  agu = queue_ptr->agranulesupper; \
  ar = queue_ptr->arodstree; \
  bgl = queue_ptr->bgranuleslower; \
  bgu = queue_ptr->bgranulesupper; \
  br = queue_ptr->brodstree; \
  }

bool before(int gamma, tempevent_type *alpha, tempevent_type *beta) 
{
  int  target, mal_target;
  unsigned int agranulesmidapprox, bgranulesmidapprox, currentqueue /*, maxqueue*/;
  tree_node_type *brodstree, *arodstree;
  queue_node_type *queue_ptr;
  unsigned int agranuleslower, agranulesupper, bgranuleslower,
               bgranulesupper, agranulesmid, bgranulesmid, arodsleft, brodsleft,
	       brodsright, arodsright;
#ifdef TESTING
  void printf();
#endif


  /* Determinate precedes? */
  if (beta->upper < alpha->lower) return FALSE;
  if (alpha->upper < beta->lower) return TRUE;

  /* Assert --> The periods of indeterminacy must overlap */

  /* Need a determinate before?*/ /* Is this true? */
  if (gamma == 100) return FALSE;

  /* Is the indeterminacy trivial? */ /* Is this true? */
  if (gamma == 1) return TRUE;

  target = pregammas[gamma];
  mal_target = pregammas[100 - gamma];

  agranuleslower = alpha->lower << 16;
  agranulesupper = (alpha->upper + 1) << 16;
  arodstree = uniform_tree;
  bgranuleslower = beta->lower << 16;
  bgranulesupper = (beta->upper + 1) << 16;
  brodstree = uniform_tree; 
  /*
  agranuleslowerboundary = TRUE;
  agranulesupperboundary = TRUE;
  agranuleslowerboundary = TRUE;
  bgranulesupperboundary = TRUE;
  */

  maxmaxqueue = (maxqueue > maxmaxqueue)? maxqueue: maxmaxqueue;
  currentqueue = 0;
  maxqueue = 0;

  do {

  /* ALPHA */
  /* Determine midpoint granule */
  agranulesmid = ((agranulesupper - agranuleslower) >> 1) + agranuleslower;
  agranulesmidapprox = agranulesmid >> 16; 

  /* BETA */
  /* Determine midpoint granule */
  bgranulesmid = ((bgranulesupper - bgranuleslower) >> 1) + bgranuleslower;
  bgranulesmidapprox = bgranulesmid >> 16; 
  /*
      arodsleft = arodstree->leftrods;
      brodsright = brodstree->rightrods;
      arodsright = arodstree->rightrods;
      brodsleft = brodstree->leftrods;
  */
  
#ifdef TESTING
  printf("gamma %d, target %d, mal_target %d\n", gamma, target, mal_target);
  printf("aglower %x, agupper %x, agmid %x, agmidapprox %x\n",
          agranuleslower, agranulesupper, agranulesmid, agranulesmidapprox);
  printf("bglower %x, bgupper %x, bgmid %x, bgmidapprox %x\n",
          bgranuleslower, bgranulesupper, bgranulesmid, bgranulesmidapprox);
#endif

  if (agranulesmidapprox < bgranulesmidapprox) {
    /* left, right */
    if ((arodstree->left) && (brodstree->right)) {
      arodsleft = arodstree->leftrods;
      brodsright = brodstree->rightrods;
      target -= arodsleft * brodsright;
      if (target <= 0) return TRUE;
      }

    /* left, left*/
    if ((arodstree->left) && (brodstree->left)) {
      if (((agranulesmid - 1) >> 16) < (bgranuleslower >> 16)) {
        brodsleft = brodstree->leftrods;
        arodsleft = arodstree->leftrods;
        target -= brodsleft * arodsleft;
        if (target <= 0) return TRUE;
        }
      else {
        PUSH(agranuleslower,agranulesmid,arodstree->left,
             bgranuleslower,bgranulesmid,brodstree->left);
        }
      }

    /* right, all*/
    if (arodstree->right) {
      /* right, right */ /*move upper off granule boundary*/
      if (((agranulesupper-1) >> 16) < bgranulesmidapprox) {
	if (brodstree->right) {
          brodsright = brodstree->rightrods;
          arodsright = arodstree->rightrods;
          target -= arodsright * brodsright;
          if (target <= 0) return TRUE;
	  }
        else {
	  /* no right for beta so push all */
          PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranuleslower,bgranulesupper,brodstree);
          }
        }
      else {
	/* right, all */
        PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranuleslower,bgranulesupper,brodstree);
	}
      }
    }

  else if (agranulesmidapprox > bgranulesmidapprox) {
    /* right, left */
    if ((brodstree->left) && (arodstree->right)) {
      brodsleft = brodstree->leftrods;
      arodsright = arodstree->rightrods;
      mal_target -= brodsleft * arodsright;
      if (mal_target <= 0) return FALSE;
      }

    /* left, left*/
    if ((brodstree->left) && (arodstree->left)) {
      if (((bgranulesmid - 1) >> 16) < (agranuleslower >> 16)) {
        brodsleft = brodstree->leftrods;
        arodsleft = arodstree->leftrods;
        mal_target -= brodsleft * arodsleft;
        if (mal_target <= 0) return FALSE;
        }
      else {
        PUSH(agranuleslower,agranulesmid,arodstree->left,
             bgranuleslower,bgranulesmid,brodstree->left);
        }
      }

    /* right, all*/
    if (brodstree->right) {
      /* right, right */ /*move upper off granule boundary*/
      if (((bgranulesupper-1) >> 16) < agranulesmidapprox) {
	if (arodstree->right) {
          arodsright = arodstree->rightrods;
          brodsright = brodstree->rightrods;
          mal_target -= brodsright * arodsright;
          if (mal_target <= 0) return FALSE;
	  }
        else {
	  /* no right for alpha so push all */
          PUSH(agranuleslower,agranulesupper,arodstree,
               bgranulesmid,bgranulesupper,brodstree->right);
          }
        }
      else {
	/* right, all */
        PUSH(agranuleslower,agranulesupper,arodstree,
             bgranulesmid,bgranulesupper,brodstree->right);
	}
      }
    }


  /* agranulesmidapprox == bgranulesmidapprox */
  else if ((!(agranulesmid & 0x0000FFFF)) && (!(bgranulesmid & 0x0000FFFF))) {
    /* both are on granule boundaries! */
    /* left, right */
    if (arodstree->left && brodstree->right) {
      arodsleft = arodstree->leftrods;
      brodsright = brodstree->rightrods;
      target -= arodsleft * brodsright;
      if (target <= 0) return TRUE;
      }
    /* right, left */
    if (arodstree->right && brodstree->left) {
      arodsright = arodstree->rightrods;
      brodsleft = brodstree->leftrods;
      mal_target -= arodsright * brodsleft;
      if (mal_target <= 0) return FALSE;
      }

    if (arodstree->left) {
      if (brodstree->left) {
	/* left, left */
        PUSH(agranuleslower,agranulesmid,arodstree->left,
             bgranuleslower,bgranulesmid,brodstree->left);
        }
      else {
	/* left, all */
	PUSH(agranuleslower,agranulesmid,arodstree->left,
             bgranuleslower,bgranulesupper,brodstree);
        }
      }
    else {
      if (brodstree->left) {
	/* all, left */
        PUSH(agranuleslower,agranulesupper,arodstree,
             bgranuleslower,bgranulesmid,brodstree->left);
        }
      else {
	/* nothing  */
        }
      }

    /* right, right */
    if (arodstree->right) {
      if (brodstree->right) {
	/* right, right */
        PUSH(agranulesmid,agranulesupper,arodstree->right,
             bgranulesmid,bgranulesupper,brodstree->right);
        }
      else {
	/* right, all */
        PUSH(agranulesmid,agranulesupper,arodstree->right,
             bgranuleslower,bgranulesupper,brodstree);
        }
      }
    else {
      if (brodstree->right) {
	/* all, right */
        PUSH(agranuleslower,agranulesupper,arodstree,
             bgranulesmid,bgranulesupper,brodstree->right);
        }
      else {
	/* nothing  */
        }
      }
    }

  else { /* one is not on a granule boundary */

    /* All in same granule? */
    if (((agranuleslower >> 16) <= ((bgranulesupper-1) >> 16)) &&
        (((agranulesupper-1) >> 16) <= ((bgranuleslower) >> 16))) /* nothing */;
    else {
      if (arodstree->left || arodstree->right)  {
        if (brodstree->left || brodstree->right)  {
	  /* left, right */
          PUSH(agranuleslower,agranulesmid,arodstree->left,
               bgranulesmid,bgranulesupper,brodstree->right);
	  /* left, left */
          PUSH(agranuleslower,agranulesmid,arodstree->left,
               bgranuleslower,bgranulesmid,brodstree->left);
	  /* right, right */
          PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranulesmid,bgranulesupper,brodstree->right);
	  /* right, left */
          PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranuleslower,bgranulesmid,brodstree->left);
          }
        else { 
	  /* left, all */
          PUSH(agranuleslower,agranulesmid,arodstree->left,
               bgranuleslower,bgranulesupper,brodstree);
	  /* right, all */
          PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranuleslower,bgranulesupper,brodstree);
          }
        }
      else {
        if (brodstree->left || brodstree->right)  {
	  /* all, left*/
          PUSH(agranuleslower,agranulesupper,arodstree,
               bgranuleslower,bgranulesmid,brodstree->left);
	  /* all, right */
          PUSH(agranuleslower,agranulesupper,arodstree,
               bgranulesmid,bgranulesupper,brodstree->right);
	  }
        else /* nothing */ ;
	}

      }
    }

  POP(agranuleslower,agranulesupper,arodstree,
      bgranuleslower,bgranulesupper,brodstree);

  } while (TRUE);
  /* SHOULD NEVER GET HERE! */
  /*printf("YABBA %d %d\n", gamma, maxqueue); */
  return FALSE;
}
    /* right, right */
    /*
    PUSH(agranulesmid,agranulesupper,arodstree->right,
            bgranulesmid,bgranulesupper,brodstree->right);
	    */
    /* left, right */
    /*
    PUSH(agranuleslower,agranulesmid,arodstree->left,
            bgranulesmid,bgranulesupper,brodstree->right);
	    */
    /* left, left */
    /*
    PUSH(agranuleslower,agranulesmid,arodstree->left,
            bgranuleslower,bgranulesmid,brodstree->left);
	    */
    /* right, left */
    /*
    PUSH(agranulesmid,agranulesupper,arodstree->right,
            bgranuleslower,bgranulesmid,brodstree->left);
	    */
