#include <stdio.h>
#include <math.h>
#include <sys/time.h>

/*
#define TESTING
*/

#ifdef TESTING
static int numoftrials = 1; /*300*/;
#else
static int numoftrials = 100; /*300*/;
#endif

#ifdef TESTING
#define PRECISION 4
#else
#define PRECISION 256
#endif
#define MULT_TABLE_MAX 16  /*if > 16, must change table*/
#define COARSENESS 0x8000
#define COARSENESS_SHIFTED 0x80000000
typedef struct {
  int lower;
  int upper;
} tempevent_type;

unsigned char mult_table[MULT_TABLE_MAX][MULT_TABLE_MAX];

typedef struct {
  int left[PRECISION+2];
  int right[PRECISION+2];
} tempdist_type;

typedef enum {FALSE, TRUE} bool;

tempdist_type uniform;
unsigned int uniform_tree_height = 0;

int maxmaxqueue = 0;
int maxqueue = 0;

int pregammas[101];

#define LEAF 0
#define INTERIOR_NODE 1

typedef struct {
  int height;
  unsigned int leftrods;
  unsigned int rightrods;
  void *left;
  void *right;
} tree_node_type;

tree_node_type spare_nodes[PRECISION*2 + 2];
int allocated = 0;
tree_node_type *uniform_tree;

typedef struct {
  unsigned int granuleslower;
  unsigned int granulesupper;
  unsigned int leftrodsadditional;
  unsigned int rightrodsadditional;
  tree_node_type *rodstree;
} queue_node_type;

queue_node_type queue_nodes[PRECISION*4 + 2];

typedef struct {
  queue_node_type *alpha;
  queue_node_type *beta;
} queue_type;

queue_type queue[2000];

tree_node_type *build_uniform_tree(int current,int height) {
  tree_node_type *node;

  node = &spare_nodes[allocated++];
  if (current == 1) {
    node->left = NULL;
    node->right = NULL;
    node->leftrods = 0;
    node->rightrods = 0;
    }
  else {
    node->left = (void *)build_uniform_tree(current >> 1,height+1);
    node->right = (void *)build_uniform_tree(current >> 1,height+1);
    node->leftrods = current >> 1;
    node->rightrods = current >> 1;
    }
  if (height > uniform_tree_height) uniform_tree_height = height;
  node->height = height;
  return node;
}

void main(argc, argv) 
  int argc;
  char *argv[]; {
  tree_node_type *build_uniform_tree();
  int i, j, k, sum;
  int times[101];
  int temps[101];
  void printf(), gettimeofday();
  bool before();
  struct timeval tp1, tp2;
  int temp1, temp2;
  int gamma, temp;
  tempevent_type alpha, beta;
  double f;

  temp = 0;
  alpha.lower = 1;
  alpha.upper = 30;
  beta.lower = 1;
  beta.upper = 30; 
  if (argc > 3) {
    alpha.lower = atoi(argv[1]);
    alpha.upper = atoi(argv[2]);
    beta.lower = atoi(argv[3]);
    beta.upper = atoi(argv[4]);
    }

  for (i=0;i<MULT_TABLE_MAX;i++) for (j=0;j<MULT_TABLE_MAX;j++) mult_table[i][j] = i*j;

#define GAMMA_MULT PRECISION * PRECISION / 100.0
  for (i=0;i<101;i++) {
    f = i;
    pregammas[i] = (int) ceil(f * GAMMA_MULT);
    /* printf("%d %d %f \n", i, pregammas[i], (i * (GAMMA_MULT)));*/
    }

  uniform_tree = build_uniform_tree(PRECISION,0);

  for (i=1+numoftrials;i<=100+numoftrials;i++) {

    gettimeofday(&tp1,NULL);

    for (j=0; j<numoftrials; j++) temp += before(i - numoftrials,&alpha,&beta);

    gettimeofday(&tp2,NULL);

    temp1 = tp1.tv_sec * 1000000 + tp1.tv_usec;
    temp2 = tp2.tv_sec * 1000000 + tp2.tv_usec;
    temp1 = temp2 - temp1;

    times[i-numoftrials] = temp1;
    temps[i-numoftrials] = temp;
    }
    for (sum=0,i = 1; i <= 100; i++) {
      sum += times[i];
#ifdef GRAPHING
      printf("%d\n", times[i]/numoftrials);
#else
      printf("%d %d %d\n", temps[i], i, times[i]/numoftrials);
#endif
      }

#ifndef GRAPHING
   printf("Average %d\n", sum/numoftrials);
   printf("m %d\n", maxmaxqueue);
#endif
}

#define HALF_PUSH_ALPHA(agl,agu,ar) \
  if (ar) {\
    queue_ptr = &queue[--currentqueue]; \
    queue_ptr->agranuleslower = agl; \
    queue_ptr->agranulesupper = agu; \
    queue_ptr->arodstree = ar; \
    }

#define HALF_PUSH_BETA(bgl,bgu,br) \
  if (br) {\
    queue_ptr = &queue[--currentqueue]; \
    queue_ptr->bgranuleslower = bgl; \
    queue_ptr->bgranulesupper = bgu; \
    queue_ptr->brodstree = br; \
    }

#define PUSH(agl,agu,ar,ala,ara,bgl,bgu,br,bla,bra) \
  a = &queue_nodes[maxnodequeue++]; \
  b = &queue_nodes[maxnodequeue++]; \
  queue_ptr = &queue[maxqueue++]; \
  queue_ptr->alpha = a; \
  queue_ptr->beta = a; \
  a->granuleslower = agl; \
  a->granulesupper = agu; \
  a->rodstree = ar; \
  a->leftrodsadditional = ala; \
  a->rightrodsadditional = ara; \
  b->granuleslower = bgl; \
  b->granulesupper = bgu; \
  b->rodstree = br; \
  b->leftrodsadditional = bla; \
  b->rightrodsadditional = bra; 

#define POP(agl,agu,ar,bgl,bgu,br) \
 if (currentqueue == maxqueue) return FALSE; \
 else { \
  queue_ptr = &queue[currentqueue++]; \
  a = queue_ptr->alpha; \
  b = queue_ptr->beta; \
  }

#define MULT(i,j) ((i < MULT_TABLE_MAX) && (j < MULT_TABLE_MAX))? mult_table[i][j]: i*j

bool before(int gamma, tempevent_type *alpha, tempevent_type *beta) 
{
  int  target, mal_target, i, j;
  tree_node_type *aleft, *aright, *bleft, *bright, *bprevious, *aprevious;
  unsigned int agranulesmidapprox, bgranulesmidapprox, currentqueue /*, maxqueue*/;
  tree_node_type *brodstree, *arodstree;
  queue_node_type *b, *a;
  queue_node_type *queue_ptr;
  unsigned int agranuleslower, agranulesupper, bgranuleslower,
               bgranulesupper, agranulesmid, bgranulesmid, arodsleft, brodsleft,
	       brodsright, arodsright;

  /* Determinate precedes? */
  if (beta->upper < alpha->lower) return FALSE;
  if (alpha->upper < beta->lower) return TRUE;

  /* Assert --> The periods of indeterminacy must overlap */

  /* Need a determinate before?*/ /* Is this true? */
  if (gamma == 100) return FALSE;

  /* Is the indeterminacy trivial? */ /* Is this true? */
  if (gamma == 1) return TRUE;

  target = pregammas[gamma];
  mal_target = pregammas[100 - gamma];

  btree = uniform_tree;
  atree = uniform_tree;

  PUSH(alpha->lower << 16,(alpha->upper + 1) << 16,uniform_tree,0,0,
       beta->lower << 16,(beta->upper + 1) << 16,uniform_tree,0,0);

  maxmaxqueue = (maxqueue > maxmaxqueue)? maxqueue: maxmaxqueue;
  currentqueue = 0;
  maxqueue = 0;

  do {

    /* ALPHA -- Determine midpoint granule */
    agranulesrange = (a->granulesupper - a->granuleslower) >> 1;
    agranulesmid = agranulesrange + a->granuleslower;
    agranulesmidapproxfloor = (agranulesmid - 1) >> 16; 
    agranulesmidapproxceil = agranulesmid >> 16; 

    /* BETA -- Determine midpoint granule */
    bgranulesrange = (b->granulesupper - b->granuleslower) >> 1;
    bgranulesmid = granulesrange + b->granuleslower;
    bgranulesmidapproxfloor = (bgranulesmid - 1) >> 16; 
    bgranulesmidapproxceil = bgranulesmid >> 16; 

    /* Mid of Alpha before Mid of Beta */
    if (agranulesmidapproxfloor < bgranulesmidapproxceil) {
      /* left, left*/
      brodsleft = brodstree->leftrods;
      arodsleft = arodstree->leftrods;
      target -= MULT(brodsleft,arodsleft);
      if (target <= 0) return TRUE;

      if (((agranulesmid - 1) >> 16) < (bgranuleslower >> 16)) {
        brodsleft = brodstree->leftrods;
        arodsleft = arodstree->leftrods;
        target -= MULT(brodsleft,arodsleft);
        if (target <= 0) return TRUE;
        }
      else {
        PUSH(agranuleslower,agranulesmid,arodstree->left,
             bgranuleslower,bgranulesmid,brodstree->left);
        }
      }

    /* right, all*/
    if (arodstree->right) {
      /* right, right */ /*move upper off granule boundary*/
      if (((agranulesupper-1) >> 16) < bgranulesmidapprox) {
	if (brodstree->right) {
          brodsright = brodstree->rightrods;
          arodsright = arodstree->rightrods;
          target -= MULT(arodsright,brodsright);
          if (target <= 0) return TRUE;
	  }
        else {
	  /* no right for beta so push all */
          HALF_PUSH_ALPHA(agranulesmid,agranulesupper,arodstree->right);
	  /*
          PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranuleslower,bgranulesupper,brodstree);
	       */
          }
        }
      else {
	/* right, all */
        HALF_PUSH_ALPHA(agranulesmid,agranulesupper,arodstree->right);
	/*
        PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranuleslower,bgranulesupper,brodstree);
	       */
	}
      }
    }

  else if (agranulesmidapprox > bgranulesmidapprox) {
    /* right, left */
    if ((brodstree->left) && (arodstree->right)) {
      brodsleft = brodstree->leftrods;
      arodsright = arodstree->rightrods;
      mal_target -= MULT(brodsleft,arodsright);
      if (mal_target <= 0) return FALSE;
      }

    /* left, left*/
    if ((brodstree->left) && (arodstree->left)) {
      if (((bgranulesmid - 1) >> 16) < (agranuleslower >> 16)) {
        brodsleft = brodstree->leftrods;
        arodsleft = arodstree->leftrods;
        mal_target -= MULT(brodsleft,arodsleft);
        if (mal_target <= 0) return FALSE;
        }
      else {
        PUSH(agranuleslower,agranulesmid,arodstree->left,
             bgranuleslower,bgranulesmid,brodstree->left);
        }
      }

    /* right, all*/
    if (brodstree->right) {
      /* right, right */ /*move upper off granule boundary*/
      if (((bgranulesupper-1) >> 16) < agranulesmidapprox) {
	if (arodstree->right) {
          arodsright = arodstree->rightrods;
          brodsright = brodstree->rightrods;
          mal_target -= MULT(brodsright,arodsright);
          if (mal_target <= 0) return FALSE;
	  }
        else {
	  /* no right for alpha so push all */
	  HALF_PUSH_BETA(bgranulesmid,bgranulesupper,brodstree->right);
	  /*
          PUSH(agranuleslower,agranulesupper,arodstree,
               bgranulesmid,bgranulesupper,brodstree->right);
	       */
          }
        }
      else {
	/* right, all */
        HALF_PUSH_BETA(bgranulesmid,bgranulesupper,brodstree->right);
	/*
        PUSH(agranuleslower,agranulesupper,arodstree,
             bgranulesmid,bgranulesupper,brodstree->right);
	     */
	}
      }
    }

  /* agranulesmidapprox == bgranulesmidapprox */
  else if ((!(agranulesmid & 0x0000FFFF)) && (!(bgranulesmid & 0x0000FFFF))) {
    /* both are on granule boundaries! */
    /* left, right */
    if (arodstree->left && brodstree->right) {
      arodsleft = arodstree->leftrods;
      brodsright = brodstree->rightrods;
      target -= MULT(arodsleft,brodsright);
      if (target <= 0) return TRUE;
      }
    /* right, left */
    if (arodstree->right && brodstree->left) {
      arodsright = arodstree->rightrods;
      brodsleft = brodstree->leftrods;
      mal_target -= MULT(arodsright,brodsleft);
      if (mal_target <= 0) return FALSE;
      }

    if (arodstree->left) {
      if (brodstree->left) {
	/* left, left */
        PUSH(agranuleslower,agranulesmid,arodstree->left,
             bgranuleslower,bgranulesmid,brodstree->left);
        }
      else {
	/* left, all */
	HALF_PUSH_ALPHA(agranuleslower,agranulesmid,arodstree->left);
	/*
	PUSH(agranuleslower,agranulesmid,arodstree->left,
             bgranuleslower,bgranulesupper,brodstree);
	     */
        }
      }
    else {
      if (brodstree->left) {
	/* all, left */
	HALF_PUSH_BETA(bgranuleslower,bgranulesmid,brodstree->left);
	/*
        PUSH(agranuleslower,agranulesupper,arodstree,
             bgranuleslower,bgranulesmid,brodstree->left);
	     */
        }
      else {
	/* nothing  */
        }
      }

    /* right, right */
    if (arodstree->right) {
      if (brodstree->right) {
	/* right, right */
        PUSH(agranulesmid,agranulesupper,arodstree->right,
             bgranulesmid,bgranulesupper,brodstree->right);
        }
      else {
	/* right, all */
        PUSH(agranulesmid,agranulesupper,arodstree->right,
             bgranuleslower,bgranulesupper,brodstree);
        }
      }
    else {
      if (brodstree->right) {
	/* all, right */
        PUSH(agranuleslower,agranulesupper,arodstree,
             bgranulesmid,bgranulesupper,brodstree->right);
        }
      else {
	/* nothing  */
        }
      }
    }

  else { /* one is not on a granule boundary */

    /* All in same granule? */
    if (((agranuleslower >> 16) <= ((bgranulesupper-1) >> 16)) &&
        (((agranulesupper-1) >> 16) <= ((bgranuleslower) >> 16))) /* nothing */;
    else {
      /* Check to see if there are leaves in alpha's tree */
      if (arodstree->left || arodstree->right)  {
        /* Check to see if there are leaves in beta's tree */
        if (brodstree->left || brodstree->right)  {
	  /* left, right */
/*	  zzzzz */
          /* left, right */
          if (arodstree->left && brodstree->right) {
            arodsleft = arodstree->leftrods;
            brodsright = brodstree->rightrods;
            target -= MULT(arodsleft,brodsright);
            if (target <= 0) return TRUE;
            }
          /* right, left */
          if (arodstree->right && brodstree->left) {
            arodsright = arodstree->rightrods;
            brodsleft = brodstree->leftrods;
            mal_target -= MULT(arodsright,brodsleft);
            if (mal_target <= 0) return FALSE;
            }

	  /* left, left */
          PUSH(agranuleslower,agranulesmid,arodstree->left,
               bgranuleslower,bgranulesmid,brodstree->left);
	  /* right, right */
          PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranulesmid,bgranulesupper,brodstree->right);
/*	  zzzzz */
	  /*
          PUSH(agranuleslower,agranulesmid,arodstree->left,
               bgranulesmid,bgranulesupper,brodstree->right);
	       */
	  /* left, left */
	  /*
          PUSH(agranuleslower,agranulesmid,arodstree->left,
               bgranuleslower,bgranulesmid,brodstree->left);
	       */
	  /* right, right */
	  /*
          PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranulesmid,bgranulesupper,brodstree->right);
	       */
	  /* right, left */
	  /*
          PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranuleslower,bgranulesmid,brodstree->left);
	       */
          }
        else {  /* no leaves for beta, just for alpha */ 
	  /* left, all */
          HALF_PUSH_ALPHA(agranuleslower,agranulesmid,arodstree->left);
	  /*
          PUSH(agranuleslower,agranulesmid,arodstree->left,
               bgranuleslower,bgranulesupper,brodstree);
	       */
	  /* right, all */
          PUSH(agranulesmid,agranulesupper,arodstree->right,
               bgranuleslower,bgranulesupper,brodstree);
          }
        }
      else {
        /* Check to see if there are leaves in beta's tree */
        if (brodstree->left || brodstree->right)  {
	  /* no leaves for alpha, just for beta */
	  /* all, left*/
	  HALF_PUSH_BETA(bgranuleslower,bgranulesmid,brodstree->left);
	  /*
          PUSH(agranuleslower,agranulesupper,arodstree,
               bgranuleslower,bgranulesmid,brodstree->left);
	       */
	  /* all, right */
          PUSH(agranuleslower,agranulesupper,arodstree,
               bgranulesmid,bgranulesupper,brodstree->right);
	  }
        else /* no leaves anywhere */ /* nothing */ ;
	}

      }
    }

  POP(agranuleslower,agranulesupper,arodstree,
      bgranuleslower,bgranulesupper,brodstree);

  } while (TRUE);
  /* SHOULD NEVER GET HERE! */
  /*printf("YABBA %d %d\n", gamma, maxqueue); */
  return FALSE;
}
    /* right, right */
    /*
    PUSH(agranulesmid,agranulesupper,arodstree->right,
            bgranulesmid,bgranulesupper,brodstree->right);
	    */
    /* left, right */
    /*
    PUSH(agranuleslower,agranulesmid,arodstree->left,
            bgranulesmid,bgranulesupper,brodstree->right);
	    */
    /* left, left */
    /*
    PUSH(agranuleslower,agranulesmid,arodstree->left,
            bgranuleslower,bgranulesmid,brodstree->left);
	    */
    /* right, left */
    /*
    PUSH(agranulesmid,agranulesupper,arodstree->right,
            bgranuleslower,bgranulesmid,brodstree->left);
	    */
/*
#define COARSENESS_IN_BITS 16
unsigned int agranulesdiff[COARSENESS_IN_BITS*2];
unsigned int bgranulesdiff[COARSENESS_IN_BITS*2];
  tempptr = &agranulesdiff[0];
  *tempptr =  (agranulesupper - agranuleslower) >> 1;
  for (i = 1; i < uniform_tree_height; i++) {
    *tempptr++ = *tempptr >> 1;
    }
  tempptr = &bgranulesdiff[0];
  *tempptr =  (bgranulesupper - bgranuleslower) >> 1;
  for (i = 1; i < uniform_tree_height; i++) {
    *tempptr++ = *tempptr >> 1;
    }
  bgranulesdiff[0] =  (bgranulesupper - bgranuleslower) >> 1;
  for (i = 1; i < uniform_tree_height; i++) {
    bgranulesdiff[i] = bgranulesdiff[i-1] >> 1;
    }
  agranulesmid = agranulesdiff[arodstree->height] + agranuleslower;
  bgranulesmid = bgranulesdiff[brodstree->height] + bgranuleslower;
*/

#ifdef JUNK
      /* Will going all the way left still be above the right guy? */
      if (agranulesmidapproxfloor < (bgranuleslower >> 16)) {
	/* Go left in Beta all the way (until a leaf) */
        for (btemp = brodstree; btemp->kind != LEAF; btemp = btemp->left) {
	  brodsright += btemp->rightrods;
          bleft = FALSE;
	  }
	/* Will going all the way right in alpha be OK? */
	if (((agranulesupper - 1) >> 16) < bgranulesmidapproxfloor) {
	  /* Go right in alpha all the way (until a leaf) */
          for (atemp = arodstree; atemp->kind != LEAF; atemp = atemp->right) {
	    arodsleft += atemp->leftrods;
            aright = FALSE;
	    }
          }
       }
      else {
        /* Go left in BETA till the mid point crosses ALPHA's midpoint */
        btemptree = brodstree; 
	brodsright += btemptree->rightrods;
	btempmid = bgranulesmid;
        for (;btemptree->kind != LEAF;) {
          bgranulesrange = bgranulesrange >> 1;
          btempmid -= bgranulesrange;
	  /* test to see if exit condition met */
	  if (agranulesmidapproxfloor < (btempmid >> 16)) break;
	  bgranulesmid = btempmid; /* establish new mid point */
	  btemptree = btemptree->left;
	  brodsright += btemptree->rightrods;
	  }
        /* left, right */
        arodsleft = arodstree->leftrods;
        target -= MULT(arodsleft,brodsright);
        if (target <= 0) return TRUE;
      }
#endif
#ifdef JUNKTESTING
  printf("gamma %d, target %d, mal_target %d\n", gamma, target, mal_target);
  printf("aglower %x, agupper %x, agmid %x, agmidapprox %x\n",
          agranuleslower, agranulesupper, agranulesmid, agranulesmidapprox);
  printf("bglower %x, bgupper %x, bgmid %x, bgmidapprox %x\n",
          bgranuleslower, bgranulesupper, bgranulesmid, bgranulesmidapprox);
#endif
