/***********************************************************************\ 
*                                                                       * 
*   File: calendric_system/newh/ucs/binding.h
*                                                                       * 
*   The MultiCal System is free software in the public domain; you can  * 
*   redistribute it and/or modify it as you wish. We ask that you       * 
*   retain credits referencing the University of Arizona and that you   * 
*   identify any changes you make.                                      * 
*                                                                       * 
*   Report problems to multical@cs.arizona.edu                          * 
*   Direct all inquiries to:    The MultiCal Project                    * 
*                               Department of Computer Science          * 
*                               University of Arizona                   * 
*                               Tucson, AZ 85721                        * 
*                               U.S.A.                                  * 
*                                                                       * 
\***********************************************************************/ 
#ifndef ucs_binding
#define ucs_binding
/* ucs_binding.h
*
*  This module implements UCS support for 
*  calendar defined function resolution and binding.
*  Routines in this module are called by the query
*  processing system when resolving calendar defined
*  functions at compile-time.
*
*/

typedef enum { arg_integer, arg_string, arg_event, arg_interval, arg_span,
	       arg_approximate_type, arg_exact_numeric_type,
	       arg_exact_integer_type
 } formal_arg_type;

typedef union {
	int integer;
	char *character;
	event_type event;
	interval_type interval;
	span_type span;
	string_struct string;
	float  float_float;
	double double_float;
	long long int long_int;
} *actual_arg_type;

/*  
*  Routine:  generic_function_call
*
*  Description:  The generic function called.  Its uses and integer passed by
*                ucs_bind_function to determine the correct function.
*
*  Arguments: f_num         --  (IN) : function number to call
*             array         --  (IN) : array of parameter pointers
*             return_val    -- (OUT) : pointer to the return value
*
*  Return Value: Error code
*
*  Errors: Depends on the function or
*          ucs_function_not_found
*    
*  Side Effects: Overwrites return_val
*/
ucs_error_type generic_function_call(int f_num, actual_arg_type *array, actual_arg_type *return_val);


/*  
*  Routine:  ucs_bind_function 
*
*  Description:  Resolve a calendric defined function name
*
*  Arguments: function_name    -- (IN) : name of the function
*             num_arguments    -- (IN) : number of arguments in the array 
*             type_array       -- (OUT) : array of function arguments codes
*             return_type_type -- (OUT): return value type for function
*             f_num            -- (OUT): handle to the function
*
*  Return Value: Error code
*
*  Errors: ucs_incompatible_types
*          ucs_function_not_found
*    
*  Side Effects: Overwrites f
*/
ucs_error_type ucs_bind_function(char *function_name, int num_arguments, formal_arg_type *type_array,
				formal_arg_type *return_type_type, int *f_num);

/*
*  Routine:  ucs_function_exists
*
*  Description:  Determines if a function exists
*
*  Arguments: function_name -- (IN) : name of the function
*
*  Return Value: bool -- TRUE if the function exists
*			 FALSE if the function does not exist
*
*  Errors:  None
*
*  Side Effects:  None
*/
bool ucs_function_exist(char *function_name);





#endif
