/***********************************************************************\ 
*                                                                       * 
*   File: calendric_system/newh/ucs/property.h
*                                                                       * 
*   The MultiCal System is free software in the public domain; you can  * 
*   redistribute it and/or modify it as you wish. We ask that you       * 
*   retain credits referencing the University of Arizona and that you   * 
*   identify any changes you make.                                      * 
*                                                                       * 
*   Report problems to multical@cs.arizona.edu                          * 
*   Direct all inquiries to:    The MultiCal Project                    * 
*                               Department of Computer Science          * 
*                               University of Arizona                   * 
*                               Tucson, AZ 85721                        * 
*                               U.S.A.                                  * 
*                                                                       * 
\***********************************************************************/ 
#ifndef ucs_property
#define ucs_property

/* ucs_property.h
*
*  This module implements UCS operations supporting 
*  property activation.  Routines in this module are called
*  by the query processing system when activating properties.
*
*/

/*  
*  Routine:  ucs_property_push
*
*  Description:  Add a property to be activated
*
*  Arguments: property -- (IN): a property
*		valud -- (IN): the value of that property
*
*  Return Value: Error code
*
*  Errors: ucs_invalid_property
*          ucs_table_overflow
*  
*  Side Effects: Adds the property to the list of properties
*		 waiting to be activated
*/
ucs_error_type ucs_property_push(char *property, char *value);

/*  
*  Routine:  ucs_property_activate
*
*  Description:  Activate any waiting properties
*
*  Arguments: None
*
*  Return Value: Error code
*
*  Errors: None
*  
*  Side Effects: Activates any pushed properties
*/
ucs_error_type ucs_property_activate();

/*  
*  Routine:  ucs_property_clear
*
*  Description:  Discard any properties waiting to be activated
*
*  Arguments: None
*
*  Return Value: Error code
*
*  Errors: None
*  
*  Side Effects: Discards any waiting properties
*/
ucs_error_type ucs_property_clear();

/*  
*  Routine:  ucs_property_deactivate
*
*  Description:  Deactivates the last set of activated properties
*
*  Arguments: None
*
*  Return Value: Error code
*
*  Errors: ucs_table_underflow
*  
*  Side Effects: Uncovers the previously active set of properties
*/
ucs_error_type ucs_property_deactivate();

/*  
*  Routine:  ucs_property_return
*
*  Description:  Finds the current value of a property
*
*  Arguments: property_num -- (IN) The number of the property being searched
*             value        -- (OUT) The place to put the property value
*
*  Return Value: Error code
*
*  Errors: ucs_invalid_property, ucs_string_overflow
*  
*  Side Effects: Overwrites the value string
*/
ucs_error_type ucs_property_return(int property_num, string_struct value);
#endif
