/***********************************************************************\ 
*                                                                       * 
*   File: calendric_system/tadt/tadt_memory.c
*                                                                       * 
*   The MultiCal System is free software in the public domain; you can  * 
*   redistribute it and/or modify it as you wish. We ask that you       * 
*   retain credits referencing the University of Arizona and that you   * 
*   identify any changes you make.                                      * 
*                                                                       * 
*   Report problems to multical@cs.arizona.edu                          * 
*   Direct all inquiries to:    The MultiCal Project                    * 
*                               Department of Computer Science          * 
*                               University of Arizona                   * 
*                               Tucson, AZ 85721                        * 
*                               U.S.A.                                  * 
*                                                                       * 
\***********************************************************************/ 
#include <stdlib.h>
#include "system_wide.h"
#include "tadt.h"

/* 
 * tadt global variables follow
 */

dist_space_type uniform;
dist_type distributions[MAX_DISTRIBUTION] = {&uniform};
seconds_type tadt_secs1;
event_type   tadt_event1;
span_type    tadt_span1;
span_type    tadt_zero_span;
unsigned long leap_second_table[MAX_LEAP_SECOND] = {0,1,2,3,4,5,6,7,8,9,10,
						   11,12,13,14,15,16,17,18,19,
						   20,21,22,23,24,25,26};
unsigned long leap_second_probe_table[MAX_LEAP_SECOND] = {0,
                           78796800,
                           78796800,
                           78796800,
                           78796800,
                           78796800,
                           78796800,
                           78796800,
                           78796800,
                           78796800,
                           78796800,
                           94694400,
                          126230400,
                          157766400,
                          189302400,
                          220924800,
                          252460800,
                          283996800,
                          315532800,
                          362793600,
                          394329600,
                          425865600,
                          489024000,
                          536457600,
                          599616000,
                          631152000,
                          678412800};

size_type sizes[MAX_RESOLUTIONS_OF_EVENTS][MAX_PRECISION];

#ifdef INDETERMINATE
span_space_type chunk32_table[HOW_MANY_CHUNKS][MAX_CHUNK];
span_space_type chunk64_table[HOW_MANY_CHUNKS][MAX_CHUNK];
span_space_type chunk96_table[HOW_MANY_CHUNKS][MAX_CHUNK];
dist_space_type          dist_table[HOW_MANY_DISTS];
#endif

/*  
*  Routine: tadt_init
*
*  Description: Allocate globals shared in the tadt system.
*               These globals are shared variables such as seconds structures.
*
*  Arguments:   None
*
*  Return Value: None
*
*  Errors: None
*  
*  Side Effects: overwrites the tadt global pointers
*/
void tadt_init()
{
  event_space_type temp_present;
  int i, j;
  int32 temp32, chunk32;
  int64 temp64, chunk64;
  int96 temp96, chunk96;

#ifdef INDETERMINATE
  int temp1;
#endif 

  tadt_present_as_of_query(TRUE, &temp_present);

  tadt_secs1 = tadt_allocate_seconds(determinate);
  tadt_event1 = tadt_allocate_e(determinate);
  tadt_span1 = tadt_allocate_s(determinate);
  tadt_zero_span = tadt_allocate_s(upack);
  tadt_zero_span->upack96.granules = 0; /* Should cover 32, 64, and 96 */

  /* Build the sizes Table */
  for (i=0; i < MAX_RESOLUTIONS_OF_EVENTS; i++) {
    for (j=0; j < MAX_PRECISION; j++) {
      sizes[i][j] = SIZE32;
      }
    }

  sizes[determinate][9] = SIZE64;
  sizes[determinate][10] = SIZE64;
  sizes[determinate][11] = SIZE64;
  sizes[determinate][12] = SIZE64;
  sizes[determinate][13] = SIZE64;
  sizes[determinate][14] = SIZE64;
  sizes[determinate][15] = SIZE64;
  sizes[determinate][16] = SIZE64;
  sizes[determinate][17] = SIZE64;
  sizes[determinate][18] = SIZE64;

  sizes[determinate][19] = SIZE96;
  sizes[determinate][20] = SIZE96;
  sizes[determinate][21] = SIZE96;
  sizes[determinate][22] = SIZE96;
  sizes[determinate][23] = SIZE96;
  sizes[determinate][24] = SIZE96;
  sizes[determinate][25] = SIZE96;
  sizes[determinate][26] = SIZE96;
  sizes[determinate][27] = SIZE96;

#ifdef INDETERMINATE

  /* Build the 32 bit Chunk Table */
  LOAD32(86400,chunk32);
  for (i=0; i < HOW_MANY_CHUNKS; i++) {
    for (j=0; j < MAX_CHUNK; j++) {
      chunk32_table[i][j].upack.size = SIZE32;
      chunk32_table[i][j].upack.ts_type = upack;
      LOAD32(i,temp32);
      MULT32(temp32,chunk32,chunk32_table[i][j].upack32.granules);
      }
      LOAD32(1,temp32);
      MULT32(chunk32,temp32,chunk32);
    }

  /* Build the 64 bit Chunk Table */
  LOAD64(1,chunk64);
  for (i=0; i < HOW_MANY_CHUNKS; i++) {
    for (j=0; j < MAX_CHUNK; j++) {
      chunk64_table[i][j].upack.size = SIZE64;
      chunk64_table[i][j].upack.ts_type = upack;
      LOAD64(i,temp64);
      MULT64(temp64,chunk64,chunk64_table[i][j].upack64.granules);
      }
    LOAD64(1,temp64);
    MULT64(chunk64,temp64,chunk64);
    }

  /* Build the 96 bit Chunk Table */
  LOAD96(1,chunk96);
  for (i=0; i < HOW_MANY_CHUNKS; i++) {
    for (j=0; j < MAX_CHUNK; j++) {
      chunk96_table[i][j].upack.size = SIZE96;
      chunk96_table[i][j].upack.ts_type = upack;
      LOAD96(i,temp96);
      MULT96(temp96,chunk96,chunk96_table[i][j].upack96.granules);
      }
    LOAD96(1,temp96);
    MULT96(chunk96,temp96,chunk96);
    }

  /* Build the Distribution Table */
  temp1 = 1;
  for (i=SIG_BITS; i>0; i--) temp1 = temp1*2;

  for (i=DIST_MIN; i<=DIST_MAX; i++) 
       dist_table[UNIFORM].floor[i] = (temp1*(i+1))/(DIST_MAX + 1);
  for (i=DIST_MIN; i<=DIST_MAX; i++) 
       dist_table[UNIFORM].ceiling[DIST_MAX - i] = (temp1*(DIST_MAX - i))/(DIST_MAX + 1);
  dist_table[UNIFORM].floor[DIST_MAX] = temp1 - 1;

  for (i=DIST_MIN; i<=DIST_MAX; i++) {
    uniform.ceiling[i] = (temp1*(i))/DIST_MAX;
    if ((temp1*(i)) % DIST_MAX) uniform.ceiling[i]++;
    uniform.floor[i] = (temp1*(i))/DIST_MAX;
    }

#endif

}  

/*  
*  Routine: tadt_malloc
*
*  Description: Allocate storage from the TADT memory heap.  Marks a region
*               of mem_size bytes which is now in use.  If no more memory
*               can be allocated, the behaviour of this function is undefined.
*               Currently, this function zero's out the allocated space, 
*               but this behavior is subject to change.
*
*  Arguments: size -- (IN): the number of bytes to allocate
*
*  Return Value: The address of allocated memory
*
*  Errors: None
*  
*  Side Effects: Some of the TADT heap is marked as in use. This may lead
*                to memory leaks later on.
*/

void *tadt_malloc(int size)
{
  int i;
  char *where, what, *asdf;
  

  what = 0;
  where = (char *)malloc(size);
  asdf = where;
  for (i = 0; i < size; i++) *where++ = what;

  return (void *)asdf;
}

/*  
*  Routine:  tadt_free
*
*  Description: Frees previously allocated storage
*
*  Arguments: address -- (IN):  address of previously allocated storage
*
*  Return Value: TRUE if the address was de-allocated, FALSE otherwise
*
*  Errors: None
*
*  Side Effects: If the address has not been allocated previously, then
*                this will could lead to internal errors.
*/

bool tadt_free(void *address)
{
#ifdef stupid_free
  return (bool)free(address);
#else
  free(address);
  return TRUE;
#endif
}

/* 
*  Routine: tadt_allocate_e
*
*  Description: Memory allocation for an event
*
*  Arguments: resolution -- (IN): which resolution to allocate
*
*  Return Value: Address of allocated event
* 
*  Errors: None
*
*  Side Effects: Mallocs space
*/

event_type tadt_allocate_e(resolution_type resolution)
{
  event_type event;

  /* Always allocate the biggest */
  event = (event_type)tadt_malloc(sizeof(event_space_type));
  event->generic.ts_type = resolution;

  return event;
}

/* 
*  Routine: tadt_allocate_s
*  
*  Description: Memory allocation for an span
*
*  Arguments: resolution -- (IN): which resolution to allocate
*
*  Return Value: Address of allocated span
* 
*  Errors: None
*
*  Side Effects: Mallocs space
*/

span_type tadt_allocate_s(resolution_type resolution)
{
  span_type span;

  /* Always allocate the biggest */
  span = (span_type)tadt_malloc(sizeof(span_space_type));
  span->generic.ts_type = resolution;

  return span;
}

/* 
*  Routine: tadt_allocate_i
*  
*  Description: Memory allocation for an interval
*
*  Arguments: resolution1 -- (IN): resolution of beginning event
*                                  resolution of the ending event
*
*  Return Value: Address of allocated interval
* 
*  Errors: None
*
*  Side Effects: Mallocs space
*/

interval_type tadt_allocate_i(resolution_type resolution1, resolution_type resolution2)
{
  interval_type interval;

  /* Always allocate the biggest */
  interval = (interval_type)tadt_malloc(sizeof(interval_space_type));
  interval->start = tadt_allocate_e(resolution1);
  interval->terminate = tadt_allocate_e(resolution2);
  return interval;
}

/* 
*  Routine: tadt_allocate_ts   
*  
*  Description: Memory allocation for an time-stamp.  Allocates the max space.
*
*  Arguments: None
*
*  Return Value: Address of allocated time-stamp
* 
*  Errors: None
*
*  Side Effects: Mallocs space
*/

timestamp_type tadt_allocate_ts()
{
  return (timestamp_type)tadt_malloc(sizeof(timestamp_space_type));
}

/* 
*  Routine: tadt_allocate_vs_struct
*  
*  Description: Memory allocation for an variable span structure.
*
*  Arguments: None
*
*  Return Value: Address of allocated vs_struct
* 
*  Errors: None
*
*  Side Effects: Mallocs space
*/

vs_type tadt_allocate_vs_struct()
{
  return (vs_type)tadt_malloc(sizeof(vs_space_type));
}

/* 
*  Routine: tadt_allocate_avg_state_e
*  
*  Description: Memory allocation for an average event state structure.
*
*  Arguments: None
*
*  Return Value: Address of allocated average event state structure
* 
*  Errors: None
*
*  Side Effects: Mallocs space
*/

avg_state_e tadt_allocate_avg_state_e()
{
  avg_state_e size;

  return (avg_state_e)tadt_malloc(sizeof(*size));
}


/* 
*  Routine: tadt_allocate_avg_state_fs
*  
*  Description: Memory allocation for an average fixed span state structure.
*
*  Arguments: None
*
*  Return Value: Address of allocated average fixed span state structure
* 
*  Errors: None
*
*  Side Effects: Mallocs space
*/

avg_state_fs tadt_allocate_avg_state_fs()
{
  avg_state_fs size;
  return (avg_state_fs)tadt_malloc(sizeof(*size));
}

/* 
*  Routine: tadt_allocate_string_struct
*  
*  Description: Memory allocation for string structure
*
*  Arguments: length -- (IN): the size of the string to allocate
*
*  Return Value: Address of allocated string structure
* 
*  Errors: None
*
*  Side Effects: Mallocs space
*/

string_struct tadt_allocate_string_struct(int capacity)
{
  string_struct temp;

  temp = (string_struct)tadt_malloc(sizeof(*temp));
  temp->capacity = capacity;
  temp->length = 0;
  temp->string = tadt_malloc(capacity+1);

  return temp;
}

/* 
*  Routine: tadt_allocate_seconds
*  
*  Description: Memory allocation for seconds structure
*
*  Arguments: resolution -- (IN): Resolution of allocated thing
*
*  Return Value: Address of allocated seconds structure
* 
*  Errors: None
*
*  Side Effects: Mallocs space
*/

seconds_type tadt_allocate_seconds(resolution_type resolution)
{
  seconds_type sec;

  sec = (seconds_type)tadt_malloc(sizeof(seconds_space_type));
  sec->resolution = resolution;
  return sec;
}
