/***********************************************************************\ 
*                                                                       * 
*   File: calendric_system/tadt/tadt_tools.c
*                                                                       * 
*   The MultiCal System is free software in the public domain; you can  * 
*   redistribute it and/or modify it as you wish. We ask that you       * 
*   retain credits referencing the University of Arizona and that you   * 
*   identify any changes you make.                                      * 
*                                                                       * 
*   Report problems to multical@cs.arizona.edu                          * 
*   Direct all inquiries to:    The MultiCal Project                    * 
*                               Department of Computer Science          * 
*                               University of Arizona                   * 
*                               Tucson, AZ 85721                        * 
*                               U.S.A.                                  * 
*                                                                       * 
\***********************************************************************/ 
#include <stdlib.h>
#include "system_wide.h"
#include "tadt.h"

void dump_word_by_word(event_type event, int words)
{
  int i;
  void printf();

  printf("hex ");
  for (i = 0; i < words; i++) printf("%x ", event->word[i]);
  printf("\n");
}

void dump_event(event_type event, char *name)
{
  void printf();
#ifdef INDETERMINATE
  event_type lower, upper;
#endif

  printf("EVENT=========%s ", name);
  switch (event->generic.ts_type) {
    case determinate:
      printf("DETERMINATE\n");
      dump_word_by_word(event,sizeof(determinate_type)/4);
      printf("ts_type %d\n", event->determinate.ts_type);
      printf("sign %d \n", event->determinate.sign);
      printf("data %d \n", event->determinate.data);
      break;

    case upack:
#ifdef INDETERMINATE
      if (event->upack.indeterminate) {
        printf("UPACK_IND\n");
        dump_word_by_word(event,sizeof(upack_indeterminate_type)/4);
        printf("indeterminate %d\n", event->upack_ind.indeterminate);
        printf("modified %d\n", event->upack_ind.modified);
        printf("base_ts_type %d\n", event->upack_ind.base_ts_type);
        printf("\nlower\n");
        dump_event(&event->upack_ind.lower,"lower");
        printf("\nupper\n");
        dump_event(&event->upack_ind.upper,"upper");
        break;
      }
#endif INDETERMINATE
      printf("UPACK\n");
      dump_word_by_word(event,sizeof(upack_determinate_type)/4);
      printf("indeterminate %d\n", event->upack.indeterminate);
      printf("modified %d\n", event->upack.modified);
      printf("base_ts_type %d\n", event->upack.base_ts_type);
      if (!indeterminate) {
	switch (event->upack.size) {
	  case SIZE32: printf("granules32 %d (hex) %x \n", event->upack32.granules, event->upack32.granules); break;
	  case SIZE64: printf("granules64 %d %d (hex) %x %x \n", event->upack64.granules, event->upack64.granules); break;
          default: internal_error("%s %d", "dump_event invalid upack size", event->generic.ts_type);
          }
        }
#ifdef INDETERMINATE
      else {
	lower = (event_type)&event->upack_ind.lower;
	upper = (event_type)&event->upack_ind.lower;
	switch (event->upack_ind.size) {
	  case SIZE32: printf("granules32 %d (hex) %x \n", lower->upack32.granules, lower->upack32.granules); 
	           printf("granules32 %d (hex) %x \n", upper->upack32.granules, upper->upack32.granules); break;
	  case SIZE64: printf("granules64 %d %d (hex) %x %x \n", lower->upack64.granules, lower->upack64.granules); 
	           printf("granules64 %d %d (hex) %x %x \n", upper->upack64.granules, upper->upack64.granules); break;
          default: internal_error("%s %d", "dump_event invalid upack size", event->generic.ts_type);
          }
        }
#endif
      break;

    case special:
      printf("vsapn %d\n", event->special.vspan);
      printf("which %d\n", event->special.which);
      break;

    default: internal_error("%s %d", "dump_event bad type ", event->generic.ts_type);
    }
}
