/***********************************************************************\ 
*                                                                       * 
*   File: calendric_system/tadt/tadt_tools.c
*                                                                       * 
*   The MultiCal System is free software in the public domain; you can  * 
*   redistribute it and/or modify it as you wish. We ask that you       * 
*   retain credits referencing the University of Arizona and that you   * 
*   identify any changes you make.                                      * 
*                                                                       * 
*   Report problems to multical@cs.arizona.edu                          * 
*   Direct all inquiries to:    The MultiCal Project                    * 
*                               Department of Computer Science          * 
*                               University of Arizona                   * 
*                               Tucson, AZ 85721                        * 
*                               U.S.A.                                  * 
*                                                                       * 
\***********************************************************************/ 
#include <stdlib.h>
#include "system_wide.h"
#include "tadt.h"

void dump_word_by_word(event_type event, int words)
{
  int i;
  void printf();

  printf("hex ");
  for (i = 0; i < words; i++) printf("%x ", event->word[i]);
  printf("\n");
}

void dump_packed_event(packed_event_type event, char *name)
{
#ifdef INDETERMINATE
  event_type lower, upper;
#endif

  printf("EVENT=========%s ", name);
  switch (event->generic.ts_type) {

    case packed_determinate:
      printf("DETERMINATE\n");
      dump_word_by_word(event,sizeof(determinate_type)/4);
      printf("ts_type %d\n", event->determinate.ts_type);
      printf("sign %d \n", event->determinate.sign);
      printf("data %d \n", event->determinate.data);
      break;

    default: printf("%s %d", "tried to dump bad packed event type ", event->generic.ts_type);
    }
  }

void dump_event(event_type event, char *name)
{
#ifdef INDETERMINATE
  event_type lower, upper;
#endif

  printf("EVENT=========%s ", name);
  switch (event->upack.ts_type) {

#ifdef INDETERMINATE
    case upackind:
      printf("UPACK_IND\n");
      dump_word_by_word(event,sizeof(upack_indeterminate_type)/4);
      printf("indeterminate \n");
      printf("modified %d\n", event->upack_ind.modified);
      printf("base_ts_type %d\n", event->upack_ind.base_ts_type);
      printf("\nlower\n");
      dump_event(&event->upack_ind.lower,"lower");
      printf("\nupper\n");
      dump_event(&event->upack_ind.upper,"upper");
      break;
#endif INDETERMINATE

    case upack32:
      printf("UPACK32\n");
      dump_word_by_word(event,sizeof(upack_determinate_type)/4);
      printf("modified %d\n", event->upack.modified);
      printf("base_ts_type %d\n", event->upack.base_ts_type);
      printf("granules32 %d (hex) %x \n", event->upack32.granules, event->upack32.granules);
      break;

    case upack64:
      printf("UPACK64\n");
      dump_word_by_word(event,sizeof(upack_determinate_type)/4);
      printf("modified %d\n", event->upack.modified);
      printf("base_ts_type %d\n", event->upack.base_ts_type);
      printf("granules64 %d %d (hex) %x %x \n", event->upack64.granules, event->upack64.granules); 
      break;

    case special:
      printf("vsapn %d\n", event->special.vspan);
      printf("which %d\n", event->special.which);
      break;

    default: internal_error("%s %d", "dump_event bad type ", event->upack.ts_type);
    }
}
